from typing import (
    cast,
    Optional,
    List,
    Dict,
    Union,
    Callable,
    Type
)
import re


from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.by import By

from quilla.ctx import Context
from quilla.common.enums import (
    ValidationTypes,
    ValidationStates,
    XPathValidationStates,
    URLValidationStates,
)
from quilla.reports import ValidationReport
from quilla.steps.base_steps import (
    BaseStepFactory,
    BaseValidation,
)


ValidationDictionary = Dict[str, Union[str, ValidationStates, ValidationTypes]]


class XPathValidation(BaseValidation):
    '''
    Class defining the behaviour for performing XPath validations

    Args:
        ctx: The runtime context for the application
        target: The XPath of the element to perform the validation against
        state: The desired state of the target web element
        driver: An optional argument to allow the driver to be bound at object creation.
    '''
    def __init__(
        self,
        ctx: Context,
        target: str,
        state: XPathValidationStates,
        parameters: Optional[Dict],
        driver: Optional[WebDriver] = None,
    ) -> None:
        selector: Dict[ValidationStates, Callable[[], ValidationReport]] = {
            XPathValidationStates.EXISTS: self._check_exists,
            XPathValidationStates.NOT_EXISTS: self._check_not_exists,
            XPathValidationStates.VISIBLE: self._check_visible,
            XPathValidationStates.NOT_VISIBLE: self._check_not_visible,
            XPathValidationStates.TEXT_MATCHES: self._check_text_matches,
            XPathValidationStates.NOT_TEXT_MATCHES: self._check_not_text_matches,
            XPathValidationStates.HAS_PROPERTY: self._check_has_property,
            XPathValidationStates.NOT_HAS_PROPERTY: self._check_not_has_property,
            XPathValidationStates.PROPERTY_HAS_VALUE: self._check_property_has_value,
            XPathValidationStates.NOT_PROPERTY_HAS_VALUE: self._check_not_property_has_value,
            XPathValidationStates.HAS_ATTRIBUTE: self._check_has_attribute,
            XPathValidationStates.NOT_HAS_ATTRIBUTE: self._check_not_has_attribute,
            XPathValidationStates.ATTRIBUTE_HAS_VALUE: self._check_attribute_has_value,
            XPathValidationStates.NOT_ATTRIBUTE_HAS_VALUE: self._check_not_attribute_has_value,
        }
        super().__init__(
            ctx,
            ValidationTypes.XPATH,
            target,
            state,
            selector,
            parameters=parameters,
            driver=driver
        )

    def _find_all(self) -> List[WebElement]:
        '''
        Proxy method to find all elements specified by the _target attribute

        Returns:
            A list of all the elements found for that specific target, searched by XPath

        Raises:
            NoDriverException: If the driver is not currently bound to this step
        '''
        return self.driver.find_elements(By.XPATH, self.target)

    def _element_text_matches_pattern(self) -> bool:
        self._verify_parameters('pattern')
        element_text = self.element.text
        pattern = self.parameters['pattern']

        return re.search(pattern, element_text) is not None

    def _element_exists(self) -> bool:
        return len(self._find_all()) > 0

    def _element_visible(self) -> bool:
        return self.element.is_displayed()

    def _element_has_property(self) -> bool:
        self._verify_parameters('name')

        return self.element.get_property(self.parameters['name']) is not None

    def _element_has_attribute(self) -> bool:
        self._verify_parameters('name')

        return self.element.get_attribute(self.parameters['name']) is not None

    def _element_check_value(self, value_fetch_fn: Callable[[str], Optional[str]]) -> bool:
        self._verify_parameters('name', 'value')

        element_value = value_fetch_fn(self.parameters['name'])

        return element_value == self.parameters['value']

    def _element_property_has_value(self) -> bool:
        return self._element_check_value(self.element.get_property)

    def _element_attribute_has_value(self) -> bool:
        return self._element_check_value(self.element.get_attribute)

    def _check_exists(self) -> ValidationReport:
        return self._create_report(
            self._element_exists()
        )

    def _check_not_exists(self) -> ValidationReport:
        return self._create_report(
            not self._element_exists
        )

    def _check_visible(self) -> ValidationReport:
        return self._create_report(
            self._element_visible()
        )

    def _check_not_visible(self) -> ValidationReport:
        return self._create_report(
            not self._element_visible
        )

    def _check_text_matches(self) -> ValidationReport:
        text_matches = self._element_text_matches_pattern()
        msg = ''

        if not text_matches:
            msg = (
                f'Element text "{self.element.text}" '
                f'does not match pattern "{self._parameters["pattern"]}"'
            )
        return self._create_report(
            text_matches,
            msg
        )

    def _check_not_text_matches(self) -> ValidationReport:
        text_matches = self._element_text_matches_pattern()
        msg = ''

        if text_matches:
            msg = (
                f'Element text "{self.element.text}" '
                f'matches pattern "{self._parameters["pattern"]}"'
            )
        return self._create_report(
            not text_matches,
            msg
        )

    def _check_has_property(self) -> ValidationReport:
        return self._create_report(
            self._element_has_property()
        )

    def _check_not_has_property(self) -> ValidationReport:
        return self._create_report(
            not self._element_has_property()
        )

    def _check_has_attribute(self) -> ValidationReport:
        return self._create_report(
            self._element_has_attribute()
        )

    def _check_not_has_attribute(self) -> ValidationReport:
        return self._create_report(
            not self._element_has_attribute()
        )

    def _check_property_has_value(self) -> ValidationReport:
        return self._create_report(
            self._element_property_has_value()
        )

    def _check_not_property_has_value(self) -> ValidationReport:
        return self._create_report(
            not self._element_property_has_value()
        )

    def _check_attribute_has_value(self) -> ValidationReport:
        return self._create_report(
            self._element_attribute_has_value()
        )

    def _check_not_attribute_has_value(self) -> ValidationReport:
        return self._create_report(
            not self._element_attribute_has_value
        )


class URLValidation(BaseValidation):
    '''
    Class defining the behaviour for performing URL validations

    Args:
        ctx: The runtime context for the application
        target: The URL to perform the validation with
        state: The desired state of the target url
        driver: An optional argument to allow the driver to be bound at object creation.
    '''
    def __init__(
        self,
        ctx: Context,
        target: str,
        state: URLValidationStates,
        parameters: Optional[Dict],
        driver: Optional[WebDriver] = None,
    ) -> None:
        selector: Dict[ValidationStates, Callable[[], ValidationReport]] = {
            URLValidationStates.CONTAINS: self._check_contains,
            URLValidationStates.NOT_CONTAINS: self._check_not_contains,
            URLValidationStates.EQUALS: self._check_equals,
            URLValidationStates.NOT_EQUALS: self._check_not_equals
        }
        super().__init__(
            ctx,
            ValidationTypes.URL,
            target,
            state,
            selector=selector,
            parameters=parameters,
            driver=driver
        )

    def perform(self) -> ValidationReport:
        '''
        Performs the correct action based on what is defined within the selector, and returns
        the resulting report produced.

        Returns:
            A report summarizing the results of the executed validation
        '''
        report = super().perform()

        if not report.success:
            report.msg = f'Expected URL: "{self._target}", Received URL "{self.url}"'

        return report

    @property
    def url(self) -> str:
        '''
        The current URL of the browser

        Raises:
            NoDriverException: If the driver is not currently bound to this step
        '''
        return self.driver.current_url

    def _check_contains(self) -> ValidationReport:
        return self._create_report(
            self.url.find(self.target) > -1
        )

    def _check_not_contains(self) -> ValidationReport:
        return self._create_report(
            self.url.find(self.target) == -1
        )

    def _check_equals(self) -> ValidationReport:
        return self._create_report(
            self.url == self.target
        )

    def _check_not_equals(self) -> ValidationReport:
        return self._create_report(
            self.url != self.target
        )


class Validation(BaseStepFactory):
    '''
    Factory class for the different validations
    '''
    validation_selector: Dict[ValidationTypes, Type[BaseValidation]] = {
        ValidationTypes.XPATH: XPathValidation,
        ValidationTypes.URL: URLValidation,
    }

    @classmethod
    def from_dict(
        cls,
        ctx: Context,
        validation_dict: ValidationDictionary,
        driver: Optional[WebDriver] = None
    ) -> BaseValidation:
        '''
        From a validation dict, produces the appropriate validation object

        Args:
            ctx: The runtime context for the application
            validation_dict: A dictionary containing the definition of a validation, including
                the target, state, and type of validation to be performed
            driver: The driver that will be connected to the validation, if any

        Returns:
            Validation object of the type requested in the validation dictionary
        '''
        validation_params = {
            'driver': driver,
            'target': validation_dict['target'],
            'state': validation_dict['state'],
            'parameters': validation_dict.get('parameters', None),
        }

        validation_type = cast(ValidationTypes, validation_dict['type'])

        validation = cls.validation_selector[validation_type]

        return validation(ctx=ctx, **validation_params)  # type: ignore



from logging import Logger
import pkg_resources
from importlib.machinery import SourceFileLoader
from pathlib import Path

import pluggy

from quilla import hookspecs


_hookimpl = pluggy.HookimplMarker('quilla')


class _DummyHooks:
    '''
    A class of dummy hook implementations that do nothing
    '''

    @_hookimpl
    def quilla_addopts():  # type: ignore
        pass

    @_hookimpl
    def quilla_context_obj():  # type: ignore
        pass

    @_hookimpl
    def quilla_configure():  # type: ignore
        pass

    @_hookimpl
    def quilla_prevalidate():  # type: ignore
        pass

    @_hookimpl
    def quilla_postvalidate():  # type: ignore
        pass

    @_hookimpl
    def quilla_step_factory_selector():  # type: ignore
        pass

    @_hookimpl
    def quilla_configure_logger():  # type: ignore
        pass


def _get_uiconf_plugins(pm: pluggy.PluginManager, root: Path, logger: Logger):
    '''
    Attempts to load a conftest.py file on the root directory, and add all defined plugin
    hooks on that file into the plugin manager

    Args:
        pm: The plugin manager
        root: The directory in which to search for the conftest file
    '''

    uiconf_file = root / 'uiconf.py'
    if not uiconf_file.exists() or not uiconf_file.is_file():
        logger.info('Could not find uiconf file in %s', uiconf_file)
        return  # No plugin file to load

    abs_path = uiconf_file.expanduser().resolve()

    uiconf_module = SourceFileLoader('uiconf', str(abs_path)).load_module()

    logger.debug('Loading hooks from "uiconf" module')
    _load_hooks_from_module(pm, uiconf_module, logger)


def _load_hooks_from_module(pm: pluggy.PluginManager, module, logger: Logger):
    '''
    Load a module into the given plugin manager object by finding all
    methods in the module that start with the `quilla_` prefix
    and wrapping the in a _hookimpl so they are picked up by `pluggy`

    Args:
        pm: The plugin manager
        module: The loaded module instance
    '''
    hooks = filter(lambda x: x.find('quilla_') == 0, dir(hookspecs))

    for hook in hooks:
        if hasattr(module, hook):
            logger.debug('Found "%s" hook in module', hook)
            hook_function = getattr(module, hook)
            hook_function = _hookimpl(hook_function)
            setattr(module, hook, hook_function)

    logger.debug('Loading all discovered hooks into the plugin manager')
    pm.register(module)


def _load_entrypoint_plugins(pm: pluggy.PluginManager, logger: Logger):
    for entry_point in pkg_resources.iter_entry_points('QuillaPlugins'):
        logger.info('Found plugin module "%s" in entrypoints', entry_point.name)
        try:
            entry_point.require()
            _load_hooks_from_module(pm, entry_point.load(), logger)
        except pkg_resources.DistributionNotFound as e:
            # Skips package if it cannot load it
            logger.info(
                'Could not find proper distributions for %s plugin module',
                entry_point.name
            )
            logger.debug('Module encountered %s', e, exc_info=True)
            pass


def get_plugin_manager(path: str, logger: Logger) -> pluggy.PluginManager:
    '''
    Creates and configures a plugin manager by loading all the plugins defined
    through entrypoints or through a `uiconf.py` file found at the `path` location

    Args:
        path: the directory in which the `uiconf.py` will be found

    Returns:
        a configured PluginManager instance with all plugins already loaded
    '''
    pm = pluggy.PluginManager('quilla')
    logger.debug('Loading dummy hooks into plugin manager')
    pm.register(_DummyHooks)

    logger.debug('Loading entrypoint plugins')
    _load_entrypoint_plugins(pm, logger)

    logger.debug('Loading local "uiconf" plugin')
    _get_uiconf_plugins(pm, Path(path), logger)

    return pm



from typing import (
    Optional,
    Dict,
    Any
)

from selenium.webdriver.remote.webdriver import WebDriver
from quilla.common.enums import UITestActions

from quilla.ctx import Context
from quilla.common.enums import (
    OutputSources
)
from quilla.steps.base_steps import (
    BaseStep,
    BaseStepFactory
)


class OutputValueStep(BaseStep, BaseStepFactory):
    required_params = [
        'target',
        'parameters'
    ]

    @classmethod
    def from_dict(
        cls,
        ctx: Context,
        action_dict: Dict,
        driver: Optional[WebDriver] = None
    ) -> "BaseStep":
        '''
        Factory method to extract needed parameters from a dictionary
        '''
        for item in cls.required_params:
            if item not in action_dict:
                raise AttributeError('Missing one or more required parameters')

        params: Dict[str, Any] = {}

        for param in cls.required_params:
            params[param] = action_dict[param]

        return OutputValueStep(ctx, **params, driver=driver)

    def __init__(
        self,
        ctx: Context,
        target: Optional[str] = None,
        parameters: Optional[dict] = None,
        driver: Optional[WebDriver] = None,
    ):
        super().__init__(ctx, UITestActions.OUTPUT_VALUE, target, parameters, driver=driver)
        self._verify_target()
        self._verify_parameters('source', 'outputName')

        self.selector = {
            OutputSources.LITERAL: self._output_literal,
            OutputSources.XPATH_TEXT: self._output_xpath_text,
            OutputSources.XPATH_PROPERTY: self._output_xpath_property,
        }

    def perform(self):
        self.ctx.logger.debug(
            'Creating value output with source %s and target %s',
            self._parameters['source'],
            self._target,
        )
        value_producer = self.selector[self.parameters['source']]

        output_value = value_producer()
        self._create_output(output_value)

    def _create_output(self, value):
        self.ctx.create_output(self.parameters['outputName'], value)

    def _output_literal(self):
        return self.target

    def _output_xpath_text(self):
        return self.element.text

    def _output_xpath_property(self):
        self._verify_parameters('propertyName')
        property_name = self.parameters['propertyName']

        return self.element.get_property(property_name)

    def copy(self) -> "OutputValueStep":
        '''
        Creates a shallow copy of the OutputValueStep object

        This is used so that each browser can have an independent copy of
        the steps, in case any script would want to edit individual browser
        steps
        '''
        return OutputValueStep(
            self.ctx,
            self._target,      # Make sure it's passed in raw
            self._parameters,  # Make sure it's passed in raw
            self._driver
        )



'''
Module for webdriver subclasses that configure the browsers
'''

from selenium import webdriver
from msedge.selenium_tools import Edge, EdgeOptions

from quilla.ctx import Context


class FirefoxBrowser(webdriver.Firefox):
    '''
    A class used to configure the Firefox browser driver for use in UIValidation module.

    Args:
        ctx: The runtime context for the application
    '''
    def __init__(self, ctx: Context):
        options = webdriver.FirefoxOptions()

        # If debugging, do not start browser in headless mode
        if ctx.run_headless:
            options.add_argument('-headless')

        super().__init__(options=options)


class ChromeBrowser(webdriver.Chrome):
    '''
    A class used to configure the Chrome browser driver for use in UIValidation module.

    Args:
        ctx: The runtime context for the application
    '''
    def __init__(self, ctx: Context):
        options = webdriver.ChromeOptions()

        if ctx.no_sandbox:
            options.add_argument('--no-sandbox')
        if ctx.run_headless:
            options.add_argument('--headless')

        super().__init__(options=options)


class EdgeBrowser(Edge):
    '''
    A class used to configure the Edge browser driver for use in UIValidation module.

    Args:
        ctx: The runtime context for the application
    '''
    def __init__(self, ctx: Context):
        options = EdgeOptions()

        options.use_chromium = True
        options.set_capability('platform', 'ANY')  # Prevent Edge from defaulting to Windows

        if ctx.no_sandbox:
            options.add_argument('--no-sandbox')
        if ctx.run_headless:
            options.add_argument('--headless')

        super().__init__(options=options)



import json

import pytest
from py._path.local import LocalPath

from quilla import (
    setup_context,
    execute
)
from quilla.reports.report_summary import ReportSummary


def collect_file(parent: pytest.Session, path: LocalPath, prefix: str):
    '''
    Collects files if their path ends with .json and starts with the prefix

    Args:
        parent: The session object performing the collection
        path: The path to the file that might be collected
        prefix: The prefix for files that should be collected

    Returns:
        A quilla file object if the path matches, None otherwise

    '''
    # TODO: change "path" to be "fspath" when pytest 6.3 is released:
    # https://docs.pytest.org/en/latest/_modules/_pytest/hookspec.html#pytest_collect_file
    if path.ext == ".json" and path.basename.startswith(prefix):
        return QuillaFile.from_parent(parent, fspath=path)


class QuillaFile(pytest.File):
    def collect(self):
        '''
        Loads the JSON test data from the path and creates the test instance

        Yields:
            A quilla item configured from the JSON data
        '''
        test_data = self.fspath.open().read()
        yield QuillaItem.from_parent(self, name=self.fspath.purebasename, test_data=test_data)


class QuillaItem(pytest.Item):
    def __init__(self, name: str, parent: QuillaFile, test_data: str):
        super(QuillaItem, self).__init__(name, parent)
        self.test_data = test_data
        json_data = json.loads(test_data)
        markers = json_data.get('markers', [])
        for marker in markers:
            self.add_marker(marker)

    def runtest(self):
        '''
        Runs the quilla test by creating an isolated context and executing the test
        data retrieved from the JSON file.
        '''
        ctx = setup_context(
            [*self.config.getoption('--quilla-opts').split(), ''],
            str(self.config.rootpath)
        )
        ctx.json = self.test_data
        results = execute(ctx)
        self.results = results
        try:
            assert results.fails == 0
            assert results.critical_failures == 0
        except AssertionError:
            raise QuillaJsonException(results)

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""
        if isinstance(excinfo.value, QuillaJsonException):
            results: ReportSummary = excinfo.value.args[0]
            return json.dumps(
                results.to_dict(),
                indent=4,
                sort_keys=True
            )
        super().repr_failure(excinfo=excinfo)

    def reportinfo(self):
        return self.fspath, 0, 'failed test: %s' % self.name


class QuillaJsonException(Exception):
    '''
    Custom exception for when Quilla files fail
    '''



# flake8: noqa

'''
All step-related classes and factories
'''


from .base_steps import (
    BaseStep,
    BaseStepFactory,
    BaseValidation,
)
from .steps import TestStep
from .outputs import OutputValueStep
from .steps_aggregator import StepsAggregator
from .validations import (
    Validation,
    XPathValidation,
    URLValidation
)



from typing import (
    List,
    Type,
    Dict
)
import json

from quilla.ctx import Context
from quilla.common.enums import (
    BrowserTargets,
    UITestActions,
    URLValidationStates,
    ValidationStates,
    ValidationTypes,
    XPathValidationStates,
    OutputSources
)
from quilla.steps.steps_aggregator import StepsAggregator
from quilla.browser.browser_validations import BrowserValidations
from quilla.reports import (
    BaseReport,
    ReportSummary,
)
from quilla.common.utils import EnumResolver


# All UI Validations
class QuillaTest(EnumResolver):
    '''
    A class to convert data into a valid QuillaTest instance, which is able to resolve
    raw text data into the appropriate enums to be used by the internal classes.

    Creates shallow copies of all the steps to ensure independence

    Args:
        ctx: The runtime context for the application
        browsers: A list of browsers to run the validations against
        root: The starting path of the validations
        setup_steps: a list of steps used to create the validation report

    Attributes:
        browsers: A list of instantiated browser validations, each containing an
            independent steps aggregator
        ctx: The runtime context for the application
    '''
    validation_type_selector: Dict[ValidationTypes, Type[ValidationStates]] = {
        ValidationTypes.XPATH: XPathValidationStates,
        ValidationTypes.URL: URLValidationStates,
    }

    @classmethod
    def from_json(cls, ctx: Context, validation_json: str) -> "QuillaTest":  # pragma: no cover
        '''
        Converts a json string into a UIValidation object
        '''
        return QuillaTest.from_dict(ctx, json.loads(validation_json))

    @classmethod
    def from_file(cls, ctx: Context, fp) -> "QuillaTest":  # pragma: no cover
        '''
        Converts an fp (a .read() supporting file-like object) containing a json
        document into a UIValidation object
        '''
        return QuillaTest.from_dict(ctx, json.load(fp))

    @classmethod
    def from_filename(cls, ctx: Context, path: str) -> "QuillaTest":  # pragma: no cover
        '''
        Reads a file at the specified path and attempts to convert it into a
        UIValidation object
        '''
        with open(path) as fp:
            return QuillaTest.from_file(ctx, fp)

    @classmethod
    def from_dict(cls, ctx: Context, validation_parameters: dict) -> "QuillaTest":
        '''
        Converts a dictionary that represents a single UIValidation test case into
        the appropriate validation object.

        Note:
            The browsers are effectively a cartesian product with the steps & validations
        '''
        root_path: str = validation_parameters['path']

        definitions = validation_parameters.get('definitions', {})

        ctx.logger.debug('Validation included definitions: %s', definitions)

        ctx.load_definitions(definitions)

        browsers: List[BrowserTargets] = []
        for browser_name in validation_parameters['targetBrowsers']:
            browsers.append(cls._name_to_enum(browser_name, BrowserTargets, ctx=ctx))

        steps = validation_parameters['steps']

        for step in steps:
            ctx.logger.debug('Processing step %s', step)

            action = step['action']
            action = cls._name_to_enum(action, UITestActions)
            step['action'] = action
            if action == UITestActions.VALIDATE:
                validation_type = step['type']
                validation_state = step['state']

                validation_type = cls._name_to_enum(validation_type, ValidationTypes, ctx=ctx)

                state_enum = cls.validation_type_selector[validation_type]

                validation_state = cls._name_to_enum(validation_state, state_enum, ctx=ctx)

                step['type'] = validation_type
                step['state'] = validation_state

            if 'parameters' in step:
                params = step['parameters']
                if 'source' in params:
                    source = params['source']
                    params['source'] = cls._name_to_enum(source, OutputSources, ctx=ctx)
                step['parameters'] = params

        return QuillaTest(
            ctx,
            browsers,
            root_path,
            steps,
        )

    def __init__(
        self,
        ctx: Context,
        browsers: List[BrowserTargets],
        root: str,
        setup_steps: list,
    ):
        self.ctx = ctx
        self._steps = steps = StepsAggregator(ctx, setup_steps)

        self.browsers: List[BrowserValidations] = []

        for browser_target in browsers:
            self.browsers.append(
                BrowserValidations(
                    ctx,
                    browser_target,
                    root,
                    steps.copy(),
                )
            )

    def validate_all(self) -> ReportSummary:
        '''
        Performs all the setup test steps required for each test case
        and executes the validations, producing a set of validation
        reports.
        '''
        validation_reports: List[BaseReport] = []
        for browser in self.browsers:
            validation_reports.extend(browser.validate())

        return ReportSummary(validation_reports)



import json
from typing import (
    Dict,
    Type,
    List,
    Callable
)

from quilla.common.enums import ReportType
from quilla.reports.base_report import BaseReport
from quilla.reports.validation_report import ValidationReport
from quilla.reports.step_failure_report import StepFailureReport


class ReportSummary:
    '''
    A class to describe a series of report objects, as well as manipulating them for test purposes.

    Args:
        reports: A list of reports to produce a summary of

    Attributes:
        reports: A list of reports used to produce a summary
        successes: The number of reports that are described as successful
        fails: The numer of reports that are not described as successful
        critical_failures: The number of reports representing critical (i.e. unrecoverable)
            failures. This will be produced at any step if it causes an exception.
        filter_by: A declarative way to filter through the various reports
    '''
    selector: Dict[str, Type[BaseReport]] = {
        'validationReport': ValidationReport,
        'stepFailureReport': StepFailureReport,
    }

    def __init__(self, reports: List[BaseReport] = []):
        self.reports = reports
        self.successes = 0
        self.fails = 0
        self.critical_failures = 0
        self.filter_by = ReportSummary.FilterTypes(self)
        self._summarize()

    def to_dict(self):
        '''
        Returns:
            a dictionary representation of the summary report
        '''
        return {
            'reportSummary': {
                'total_reports': len(self.reports),
                'successes': self.successes,
                'failures': self.fails,
                'critical_failures': self.critical_failures,
                'reports': [
                    report.to_dict() for report in self.reports
                ]
            }
        }

    def to_json(self) -> str:
        '''
        Returns:
            a json string representation of the summary report
        '''
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, summary_dict):
        '''
        Loads a ReportSummary object that is represented as a dictionary. It does not trust the
        metadata that is in the report, and will regenerate the metadata itself.
        '''
        reports = summary_dict['reportSummary']['reports']
        obj_reports = []
        for report in reports:
            # Each report has a report tag as the root of the json document
            report_type = list(report.keys())[0]
            report_object = cls.selector[report_type]
            obj_reports.append(report_object.from_dict(report))
        obj_reports = [ValidationReport.from_dict(report) for report in reports]
        return ReportSummary(obj_reports)

    @classmethod
    def from_json(cls, summary_json):
        '''
        Loads a ReportSummary object that is represented as a valid json string. Calls from_dict
        with the default json loader
        '''
        return cls.from_dict(json.loads(summary_json))

    def _summarize(self):
        '''
        Performs the summary operation over the stored reports
        '''
        for report in self.reports:
            if report.report_type == ReportType.VALIDATION:
                if report.success:
                    self.successes += 1
                else:
                    self.fails += 1
            elif report.report_type == ReportType.STEP_FAILURE:
                self.fails += 1
                self.critical_failures += 1

    class FilterTypes:
        '''
        Inner class used to provide declarative filtering syntax for ReportSummary objects.
        For example, to filter by only successful reports you would call
        `reports.filter_by.success()`
        '''
        def __init__(self, summary: "ReportSummary"):
            self._summary = summary

        def _filter(self, condition: Callable[[BaseReport], bool]) -> "ReportSummary":
            '''
            Returns a new summary with only reports that match the condition passed as
            a lambda function parameter
            '''
            reports = self._summary.reports.copy()
            filtered_reports = filter(condition, reports)

            return ReportSummary(list(filtered_reports))

        def state(self, state: str) -> "ReportSummary":
            '''
            Returns:
                a new summary with only reports that have a state matching the one
                given by the state parameter
            '''
            return self._filter(
                lambda x: isinstance(x, ValidationReport) and x.state.lower() == state
            )

        def browser(self, browser: str) -> "ReportSummary":
            '''
            Returns:
                a new summary with only reports that have a browser matching the one
                given by the browser parameter
            '''
            return self._filter(lambda x: x.browser.lower() == browser.lower())

        def successful(self) -> "ReportSummary":
            '''
            Returns:
                a new summary with only the reports that produced a success
            '''
            return self._filter(lambda x: isinstance(x, ValidationReport) and x.success)

        def failure(self) -> "ReportSummary":
            '''
            Returns:
                a new summary with only the reports that produced a failure
            '''
            return self._filter(lambda x: isinstance(x, ValidationReport) and not x.success)

        def type(self, validation_type: str) -> "ReportSummary":
            '''
            Returns:
                a new summary with only reports that have a type matching the one
                given by the type parameter
            '''
            return self._filter(
                lambda x:
                    isinstance(x, ValidationReport) and
                    x.validation_type.lower() == validation_type.lower()
            )

        def target(self, target: str) -> "ReportSummary":
            '''
            Returns:
                a new summary with only reports that have a target matching the one
                given by the target parameter
            '''
            return self._filter(
                lambda x: isinstance(x, ValidationReport) and x.target.lower() == target.lower()
            )

        def critical_failure(self) -> "ReportSummary":
            '''
            Returns:
                a new summary with only reports that constitute a critical failure
            '''

            return self._filter(lambda x: x.report_type == ReportType.STEP_FAILURE)



'''
Module containing all the requisite classes to perform test steps.

Adding new actions
-------------------

Creating new simple actions in the code is designed to be fairly straightforward, and only
requires three steps:

1. Add an entry for the action on the ``enums`` module
2. Create a function to perform the actual step under the ``TestStep`` class
3. Add an entry to the selector with the enum as a key and the function as a value

Keep in mind that the step function should also validate any required data, and that
updating the schema for proper json validation is essential.

If the parameters for the new action are expected to be enums, you must also add the logic
for converting the parameter from string to enum in the ``UIValidation`` class.
'''


from typing import (
    Optional,
    Dict,
    Any,
)

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.action_chains import ActionChains

from quilla.ctx import Context
from quilla.common.enums import (
    UITestActions,
)
from quilla.steps.base_steps import (
    BaseStepFactory,
    BaseStep
)


# Steps classes
class TestStep(BaseStep, BaseStepFactory):
    '''
    Class that contains the definition of a single test step.
    Used for setting up validations

    Args:
        ctx: The runtime context of the application
        action: The action enum for this step
        target: What the target for this step is, if applicable
        parameters: Extra options for certain actions
        aggregator: The parent object holding this step
        driver: The browser driver

    Attributes:
        selector: A dictionary that maps action enums to the action function
    '''
    required_params = [
        'action',
    ]
    optional_params = [
        'target',
        'parameters',
    ]

    @classmethod
    def from_dict(
        cls,
        ctx: Context,
        action_dict,
        driver: Optional[WebDriver] = None
    ) -> "TestStep":
        '''
        Factory method to extract needed parameters from a dictionary
        '''
        for item in cls.required_params:
            if item not in action_dict:
                raise AttributeError('Missing one or more required parameters')

        params: Dict[str, Any] = {}

        for param in cls.required_params:
            params[param] = action_dict[param]

        for param in cls.optional_params:
            if param in action_dict:
                params[param] = action_dict[param]

        return TestStep(ctx, **params, driver=driver)

    def __init__(
        self,
        ctx: Context,
        action: UITestActions,
        target: Optional[str] = None,
        parameters: Optional[dict] = None,
        driver: Optional[WebDriver] = None,
    ):
        super().__init__(ctx, action, target=target, parameters=parameters, driver=driver)
        self.selector = {
            UITestActions.CLICK: self._click,
            UITestActions.CLEAR: self._clear,
            UITestActions.SEND_KEYS: self._send_keys,
            UITestActions.NAVIGATE_TO: self._navigate_to,
            UITestActions.WAIT_FOR_VISIBILITY: self._wait_for_visibility,
            UITestActions.WAIT_FOR_EXISTENCE: self._wait_for_existence,
            UITestActions.NAVIGATE_BACK: self._navigate_back,
            UITestActions.NAVIGATE_FORWARD: self._navigate_forward,
            UITestActions.HOVER: self._hover,
            UITestActions.REFRESH: self._refresh,
            UITestActions.SET_BROWSER_SIZE: self._set_browser_size,
            UITestActions.ADD_COOKIES: self._add_cookies,
            UITestActions.SET_COOKIES: self._set_cookies,
            UITestActions.CLEAR_COOKIES: self._clear_cookies,
            UITestActions.REMOVE_COOKIE: self._remove_cookie,
        }

    def copy(self) -> "TestStep":
        '''
        Creates a shallow copy of the TestStep object

        This is used so that each browser can have an independent copy of
        the steps, in case any script would want to edit individual browser
        steps
        '''
        return TestStep(
            self.ctx,
            self.action,
            self._target,      # Make sure it's passed in raw
            self._parameters,  # Make sure it's passed in raw
            self._driver
        )

    def perform(self):
        '''
        Runs the specified action. Wrapper for selecting proper inner method
        '''
        perform_action = self.selector[self.action]

        return perform_action()

    def _click(self):
        self._verify_target()
        self.element.click()

    def _clear(self):
        self._verify_target()
        self.element.clear()

    def _send_keys(self):
        self._verify_parameters('data')
        self.element.send_keys(self.parameters['data'])

    def _navigate_to(self):
        self._verify_target()
        self.driver.get(self.target)

    def _wait_for(self, condition):
        self._verify_parameters('timeoutInSeconds')
        WebDriverWait(self.driver, self.parameters['timeoutInSeconds']).until(condition)

    def _wait_for_visibility(self):
        self._verify_target()
        self._wait_for(EC.visibility_of_element_located(self.locator))

    def _wait_for_existence(self):
        self._verify_target()
        self._wait_for(EC.presence_of_element_located(self.locator))

    def _navigate_back(self):
        self.driver.back()

    def _navigate_forward(self):
        self.driver.forward()

    def _refresh(self):
        self.driver.refresh()

    def _set_browser_size(self):
        self._verify_parameters('width', 'height')
        width = self._parameters['width']
        height = self._parameters['height']
        self.driver.set_window_size(width, height)

    def _set_cookies(self):
        self._clear_cookies()
        self._add_cookies()

    def _add_cookies(self):
        self._verify_parameters('cookieJar')
        self.driver.add_cookie(self.parameters['cookieJar'])

    def _remove_cookie(self):
        self._verify_parameters('cookieName')
        self.driver.delete_cookie(self.parameters['cookieName'])

    def _clear_cookies(self):
        self.driver.delete_all_cookies()

    def _hover(self):
        self._verify_target()
        ActionChains(self.driver).move_to_element(self.element).perform()

    def _set_zoom_level(self):
        self._verify_parameters('zoomLevel')
        zoom_level = self._parameters['zoomLevel']
        self.driver.execute_script(f'document.body.style.zoom="{zoom_level}%"')



'''
Entrypoint code for the Quilla module. Deals with setting up the parser and
the runtime context for the application, then executing the rest of the application
'''
import argparse
import sys
import json
from typing import (
    List,
)
import logging

from quilla.ui_validation import QuillaTest
from quilla.ctx import (
    Context,
    get_default_context
)
from quilla.reports import ReportSummary
from quilla.plugins import get_plugin_manager


def make_parser() -> argparse.ArgumentParser:  # pragma: no cover
    '''
    Creates the required parser to run Quilla from the command line


    Returns:
        A pre-configured ArgParser instance
    '''
    parser = argparse.ArgumentParser(
        prog='quilla',
        description='''
        Program to provide a report of UI validations given a json representation
        of the validations or given the filename containing a json document describing
        the validations
        ''',
    )

    parser.add_argument(
        '-f',
        '--file',
        dest='is_file',
        action='store_true',
        help='Whether to treat the argument as raw json or as a file',
    )
    parser.add_argument(
        'json',
        help='The json file name or raw json string',
    )
    parser.add_argument(
        '--debug',
        action='store_true',
        help="Enable debug mode",
    )
    parser.add_argument(
        '--driver-dir',
        dest="drivers_path",
        action='store',
        default='.',
        help='The directory where browser drivers are stored',
    )
    parser.add_argument(
        '-P',
        '--pretty',
        action='store_true',
        help='Set this flag to have the output be pretty-printed'
    )
    parser.add_argument(
        '--no-sandbox',
        dest='no_sandbox',
        action='store_true',
        help='''
        Adds \'--no-sandbox\' to the Chrome and Edge browsers.
        Useful for running in docker containers'
        '''
    )
    parser.add_argument(
        '-d',
        '--definitions',
        action='append',
        metavar='file',
        help='A file with definitions for the \'Definitions\' context object'
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        help='Flag to increase the verbosity of the outputs. '
        'Log outputs are directed to stderr by default.',
        default=0
    )

    return parser


def get_early_configs(parser: argparse.ArgumentParser, args: List[str]) -> argparse.Namespace:
    '''
    Extracts known configs early on. This will disable help/usage printing and prevent
    exiting on failure, since it potentially needs to ignore errors to collect the early configs

    Args:
        parser: A parser instance, such as the one returned by `make parser`
        args: A list of strings containing the arguments to be consumed

    Returns:
        A namespace object containing the early configurations for Quilla

    '''

    # Saves functions that need to be disabled for early configs
    exit_function = parser.exit
    print_help_fn = parser.print_help
    print_usage_fn = parser.print_usage

    # "Disable" the above functions
    do_nothing = lambda *x: None  # noqa: E731
    parser.exit = do_nothing  # type: ignore
    parser.print_help = do_nothing  # type: ignore
    parser.print_usage = do_nothing  # type: ignore

    # Get early args
    parsed_args, _ = parser.parse_known_intermixed_args(args)

    # Revert the disabled functions to their normal usage
    parser.exit = exit_function  # type: ignore
    parser.print_help = print_help_fn  # type: ignore
    parser.print_usage = print_usage_fn  # type: ignore

    return parsed_args


def make_default_logger(early_configs: argparse.Namespace) -> logging.Logger:
    '''
    Creates a default logger class for Quilla. All internal logger setup logic
    is done in this function

    Args:
        early_configs: The extracted configurations from an early pass-through
            of the parse_known_intermixed_args function, before plugins run

    Returns:
        A configured ``Logger``
    '''
    logger = logging.getLogger('quilla')

    logger.propagate = False  # Disable the root logger

    formatter = logging.Formatter(
        fmt='[%(asctime)s] [%(levelname)s] - %(message)s',
        datefmt='%H:%M:%S'
    )

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)

    logger.addHandler(stream_handler)

    if early_configs.verbose >= 2:
        logger.setLevel(logging.DEBUG)
    elif early_configs.verbose == 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    if early_configs.debug:
        logger.setLevel(logging.DEBUG)

    return logger


def execute(ctx: Context) -> ReportSummary:
    '''
    Runs all defined UI validations from the json, either from file or from raw json text, and
    prints all the resulting reports to the console

    Args:
        json: The json string describing the validation

    Returns:
        A summary of all reports produced by Quilla
    '''

    ctx.logger.debug('Building UIValidation object from JSON')
    quilla_test = QuillaTest.from_json(ctx, ctx.json)

    ctx.logger.info('Passing UIValidation instance to "quilla_prevalidate" hook')
    ctx.pm.hook.quilla_prevalidate(validation=quilla_test)

    ctx.logger.debug('Running all validations')
    reports = quilla_test.validate_all()

    ctx.logger.info('Running "quilla_postvalidate" hooks')
    ctx.pm.hook.quilla_postvalidate(ctx=ctx, reports=reports)

    return reports


def setup_context(args: List[str], plugin_root: str = '.') -> Context:
    '''
    Starts up the plugin manager, creates parser, parses args and sets up the application context

    Args:
        args: A list of cli options, such as sys.argv[1:]
        plugin_root: The directory used by the plugin manager to search for `uiconf.py` files
    Returns:
        A runtime context configured by the hooks and the args
    '''
    parser = make_parser()
    early_configs = get_early_configs(parser, args)
    logger = make_default_logger(early_configs)

    logger.debug('Default logger configured, running Quilla with args %s', args)

    logger.debug('Initializing plugin manager')

    pm = get_plugin_manager(plugin_root, logger)

    logger.info('Running "quilla_configure_logger" hook')
    pm.hook.quilla_configure_logger(logger=logger)

    logger.info('Running "quilla_addopts" hook')
    pm.hook.quilla_addopts(parser=parser)  # type: ignore

    logging.debug('Parsing all args')
    parsed_args = parser.parse_args(args)

    # Set to empty list since argparse defaults to None
    if not parsed_args.definitions:
        parsed_args.definitions = []

    if not parsed_args.is_file:
        json_data = parsed_args.json
    else:
        with open(parsed_args.json) as f:
            json_data = f.read()

    logger.debug('Initializing context object')

    ctx = get_default_context(
        pm,
        parsed_args.debug,
        parsed_args.drivers_path,
        parsed_args.pretty,
        json_data,
        parsed_args.is_file,
        parsed_args.no_sandbox,
        parsed_args.definitions,
        logger=logger,
    )

    logger.info('Running "quilla_configure" hook')
    pm.hook.quilla_configure(ctx=ctx, args=args)

    return ctx


def run():
    '''
    Creates the parser object, parses the command-line arguments, and runs them, finishing with the
    appropriate exit code.
    '''
    ctx = setup_context(sys.argv[1:])

    ctx.logger.debug('Context setup complete, running the "execute" function')

    reports = execute(ctx)

    ctx.logger.debug('Finished generating reports')

    out = reports.to_dict()
    if ctx._context_data['Outputs']:
        out['Outputs'] = ctx._context_data['Outputs']

    if ctx.pretty:
        print(json.dumps(
            out,
            indent=ctx.pretty_print_indent,
            sort_keys=True
        ))
    else:
        print(json.dumps(out))

    if reports.fails > 0:
        exit_code = 1
    else:
        exit_code = 0

    sys.exit(exit_code)


if __name__ == '__main__':
    run()



from quilla.common.enums import (
    ReportType,
    UITestActions,
)
from quilla.reports.base_report import BaseReport


class ValidationReport(BaseReport):
    '''
    Data class for producing reports on the results of the validations

    Args:
        validation_type: The string representation of the type of validation performed
        target: The validation target
        state: The desired state used for validation
        browser_name: The name of the browser that the validation was performed on
        success: Whether the validation passed or not
        msg: An optional string adding further context to the report

    Attributes:
        validation_type: The string representation of the type of validation performed
        target: The validation target
        state: The desired state used for validation
        success: Whether the validation passed or not
        msg: An optional string adding further context to the report
    '''

    def __init__(
        self,
        validation_type: str,
        target: str,
        state: str,
        browser_name: str,
        success: bool,
        msg: str = ""
    ):
        super().__init__(
            ReportType.VALIDATION,
            browser_name,
            UITestActions.VALIDATE,
            msg
        )
        self.validation_type = validation_type
        self.target = target
        self.state = state
        self.success = success

    @classmethod
    def from_dict(cls, report) -> "ValidationReport":
        '''
        Converts a dictionary into a ValidationReport object

        Args:
            report:
        '''
        params = report['validationReport']
        msg = ""
        if 'msg' in params:
            msg = params['msg']
        return ValidationReport(
            validation_type=params['type'],
            target=params['target'],
            state=params['state'],
            browser_name=params['targetBrowser'],
            success=params['passed'],
            msg=msg
        )

    def to_dict(self):
        '''
        Returns a dictionary representation of the object
        '''
        report = {
            "validationReport": {
                "action": self.action.value,
                "type": self.validation_type,
                "target": self.target,
                "state": self.state,
                "targetBrowser": self.browser,
                "passed": self.success,
            }
        }

        if self.msg:
            report['validationReport']['msg'] = self.msg

        return report



import json
from abc import (
    abstractclassmethod,
    abstractmethod,
)
from typing import Dict

from quilla.common.utils import EnumResolver
from quilla.common.enums import (
    ReportType,
    UITestActions,
)


class BaseReport(EnumResolver):
    '''
    Data class for producing reports on various steps performed

    Args:
        report_type: An enum specifying the kind of report
        browser: The name of the browser
        action: An enum specifying the kind of action that was taken
        msg: A string giving further context to the report
    '''

    def __init__(self, report_type: ReportType, browser: str, action: UITestActions, msg: str = ""):
        self.browser: str = browser
        self.action: UITestActions = action
        self.msg: str = msg
        self.report_type: ReportType = report_type

    @abstractclassmethod
    def from_dict(cls, report: Dict[str, Dict[str, str]]) -> "BaseReport":
        '''
        Converts a dictionary report into a valid Report object

        Args:
            report: a dictionary that describes a report
        '''

    @abstractmethod
    def to_dict(self):
        '''
        Converts the Report object into a dictionary representation
        '''

    def to_json(self) -> str:
        '''
        Returns:
            a json representation of the object. Good for passing reports to different applications
        '''
        report = self.to_dict()
        return json.dumps(report)

    @classmethod
    def from_json(cls, report_json: str) -> "BaseReport":
        '''
        Loads a valid json string and attempts to convert it into a Report object

        Returns:
            a report of the appropriate type
        '''
        st = json.loads(report_json)
        return cls.from_dict(st)  # type: ignore

    @classmethod
    def from_file(cls, fp) -> "BaseReport":
        '''
        Converts a fp (a file-like .read() supporting object) containing a json document
        into a Report object

        Returns:
            a report of the appropriate type
        '''
        st = json.load(fp)
        return cls.from_dict(st)  # type: ignore



'''
All shared bits of code that are used throughout much of the UIValidation package
that don't have any better place to be. This encompasses all shared utility code
such as the utility classes, enums, and exceptions used by this application.
'''



'''
Hooks that allow extension of the behaviour of steps, including output and validation
related steps. This also includes any hooks related to context expressions
'''
from typing import (
    Type,
    Dict,
    Tuple,
    Optional
)

from quilla.hookspecs import hookspec
from quilla.ctx import Context
from quilla.common.enums import UITestActions
from quilla.steps.base_steps import BaseStepFactory


StepFactorySelector = Dict[UITestActions, Type[BaseStepFactory]]


@hookspec
def quilla_step_factory_selector(selector: StepFactorySelector):
    '''
    A hook called immediately before resolving the step factory for a given step definition.
    This is used to register new step factories for custom step objects.

    Most custom steps should just add themselves to one of the non-factory selector hooks, but if
    a custom step requires complex logic it might be beneficial to register a factory to
    have more fine-grained control over the logic

    Args:
        selector: The factory selector dictionary.
    '''


@hookspec(firstresult=True)
def quilla_context_obj(ctx: Context, root: str, path: Tuple[str]) -> Optional[str]:
    '''
    A hook to allow pluggins to resolve a context object given its root
    and a path. All plugins that implement this hook *must* return None if they cannot
    resolve the context object.

    It is not possible to override the default context object handlers

    Args:
        ctx: The runtime context for the application
        root: The name of the context object, which is expressed as the root
            of a dot-separated path in the validation files
        path: The remainder of the context object path, where data is being
            retrieved from

    Returns
        the data stored at the context object if existing, None otherwise
    '''



# flake8: noqa
'''
All report generation and summary classes
'''


from .base_report import BaseReport
from .validation_report import ValidationReport
from .step_failure_report import StepFailureReport
from .report_summary import ReportSummary



import pytest
from _pytest.config import Config
from _pytest.config.argparsing import Parser

from pytest_quilla.pytest_classes import collect_file


def pytest_addoption(parser: Parser):
    '''
    Adds quilla INI option for enabling
    '''
    parser.addini(
        'use-quilla',
        'Enables or disables the use of Quilla for pytest',
        type='bool',
        default=False
    )
    parser.addini(
        'quilla-prefix',
        'Prefix that JSON files should have to be considered quilla test files',
        type='string',
        default='quilla'
    )


def pytest_load_initial_conftests(early_config: Config, parser: Parser):
    if not early_config.getini('use-quilla'):
        early_config.pluginmanager.set_blocked('quilla')
        return

    parser.addoption(
        "--quilla-opts",
        action="store",
        default="",
        help="Options to be passed through to the quilla runtime for the scenario tests"
    )


def pytest_collect_file(parent: pytest.Session, path):
    return collect_file(parent, path, parent.config.getini('quilla-prefix'))



import os
import re
from functools import lru_cache
from typing import (
    Optional,
    List,
    Dict,
    cast
)
from pathlib import Path
from logging import (
    Logger,
    getLogger,
    NullHandler,
)
import json

from pluggy import PluginManager
import pydeepmerge as pdm

from quilla.common.exceptions import (
    InvalidContextExpressionException,
    InvalidOutputName,
)
from quilla.common.utils import DriverHolder


class Context(DriverHolder):
    '''
    Class defining configurations for the runtime context. This object
    should not be created directly but retrieved with "get_default_context"

    Args:
        plugin_manager: A configured PluginManager instance
        debug: Whether the configurations should be run as debug mode.
        drivers_path: The directory where the different browser drivers are stored
        pretty: If the output should be pretty-printed
        json_data: The json describing the validations
        is_file: Whether a file was originally passed in or if the raw json was passed in
        no_sandbox: Whether to pass the '--no-sandbox' arg to Chrome and Edge
        logger: An optional configured logger instance.


    Attributes:
        pm: A PluginManager instance with all hooks already loaded
        suppress_exceptions: Whether to suppress exceptions by generating reports or
            to crash the application on exception
        run_headless: Whether the browsers should be instructed to run headless
        close_browser: Whether the cleanup process should close browsers or leave the
            session open
        pretty: If the output should be pretty-printed
        json_data: The json describing the validations
        is_file: Whether a file was originally passed in or if raw json was passed in
        no_sandbox: Whether to pass the '--no-sandbox' arg to Chrome and Edge
        logger: A logger instance. If None was passed in for the 'logger' argument, will create
            one with the default logger.
    '''
    default_context: Optional["Context"] = None
    _drivers_path: str
    _expression_regex = re.compile(r'\${{(.*)}}')
    _context_obj_expression = re.compile(
        # Used on the inside of the _expession_regex to
        # find context objects embedded into the
        # context expression regex
        r'([a-zA-Z][a-zA-Z0-9_]+)(\.[a-zA-Z_][a-zA-Z0-9_]+)+'
    )
    _output_browser: str = 'Firefox'
    pretty_print_indent: int = 4

    def __init__(
        self,
        plugin_manager: PluginManager,
        debug: bool = False,
        drivers_path: str = '.',
        pretty: bool = False,
        json_data: str = '',
        is_file: bool = False,
        no_sandbox: bool = False,
        definitions: List[str] = [],
        logger: Optional[Logger] = None
    ):
        super().__init__()
        self.pm = plugin_manager
        self._path = os.environ['PATH']  # snapshot the path
        self.is_debug = debug
        self.pretty = pretty
        self.json = json_data
        self.is_file = is_file
        self.no_sandbox = no_sandbox
        path = Path(drivers_path)

        if logger is None:
            self.logger = getLogger('quilla')
            self.logger.addHandler(NullHandler())
        else:
            self.logger = logger

        self.drivers_path = str(path.resolve())
        self._context_data: Dict[str, dict] = {'Validation': {}, 'Outputs': {}, 'Definitions': {}}
        self._load_definition_files(definitions)

    @property
    def is_debug(self) -> bool:
        '''
        A set of debug configurations. Will return true if 'debug' is originally passed or
        this property is set, but one could edit the individual permissions to make a more
        fine-tuned debugging experience
        '''

        return self._debug

    @is_debug.setter
    def is_debug(self, v: bool):
        self._debug = v
        self.suppress_exceptions: bool = not v
        self.run_headless: bool = not v
        self.close_browser: bool = not v

    @property
    def drivers_path(self) -> str:
        '''
        The path where the drivers will be stored. Setting this property will add it to
        the PATH variable, so if the drivers are already in the PATH this can be omitted.
        '''
        return self._drivers_path

    @drivers_path.setter
    def drivers_path(self, v: str) -> str:
        path = Path(v)
        driver_path_str = str(path.resolve())
        self.logger.debug('Setting driver path to "%s"', driver_path_str)
        self._drivers_path = driver_path_str
        self._set_path()

        return v

    def _set_path(self):
        os.environ['PATH'] = f"{self._path}:{self._drivers_path}"

    @lru_cache
    def perform_replacements(self, text: str) -> str:
        '''
        Extracts any relevant context expressions from the text and attempts to
        making suitable replacements for the context objects

        Args:
            text: Any string that supports context expressions

        Returns:
            The resulting string after executing the context expression

        Examples:
            >>> from quilla.ctx import Context
            >>> ctx = Context(context_data={'name': 'examplesvc'})
            >>> ctx.perform_replacements('/api/${{ Validation.name }}/get')
            '/api/examplesvc/get'
        '''
        self.logger.debug('Performing replacement function on %s', text)
        while (expression_match := self._expression_regex.search(text)) is not None:
            expression = expression_match.group(1).strip()
            self.logger.debug('Found expression match: %s', expression)
            processed = self._process_objects(expression)

            self.logger.debug('Post-processing, expression value is: %s', processed)

            text = (
                text[:expression_match.start()] +
                f'{processed}' +
                text[expression_match.end():]
            )
        return text

    def _escape_quotes(self, text: str) -> str:
        return text.replace("'", "\\'")

    def _process_objects(self, expression: str) -> str:
        '''
        Performs context object replacement to ensure that the returned string is ready for
        the eval that will occur in perform_replacements

        Args:
            expression: a string that matches the _expression_regex regular expression

        Returns:
            an eval-ready string wherein all the valid context objects have been replaced
            with the appropriate values from their respective sources
        '''
        while (object_match := self._context_obj_expression.search(expression)) is not None:
            object_expression = object_match.group(0)  # Grab the full expression

            root, *path = object_expression.split('.')
            repl_value = ''

            if root == 'Environment':
                repl_value = self._escape_quotes(os.environ.get('.'.join(path), ''))
            elif root == 'Validation' or root == 'Definitions':
                data = self._context_data[root]
                data = self._walk_data_tree(data, path, object_expression)

                repl_value = cast(str, data)
            elif self.pm is not None:
                # Pass it to the defined hooks

                self.logger.debug(
                    'Context object "%s" does not match known options, forwarding to plugins',
                    root
                )

                hook_results = self.pm.hook.quilla_context_obj(
                    ctx=self,
                    root=root,
                    path=tuple(path)
                )  # type: ignore

                # Hook results will always be either size 1 or 0
                if len(hook_results) == 0:
                    repl_value = ''
                else:
                    repl_value = hook_results[0]

            if repl_value == '':
                self.logger.info(
                    'Context expression "%s" does not resolve to any value',
                    object_expression
                )

            expression = (
                expression[:object_match.start()] +
                f'{repl_value}' +
                expression[object_match.end():]
            )
        return expression

    def _walk_data_tree(self, data, exp, object_expression):
        for entry in exp:
            data = data.get(entry, None)
            if data is None:
                raise InvalidContextExpressionException(
                    f'\'{object_expression}\' does not exist'
                )
        return data

    def create_output(self, output_name: str, value: str):
        '''
        Creates an output on the context object to be returned by the validation
        module if the browser is the supported output browser, and sets the value
        in the Validation context object. Setting the value in the
        Validation context happens regardless of the browser type, since further
        steps on the validation chain could need this specific value and the order in
        which browsers are executed is dependent on the order that the user gave
        in the validation json.

        Args:
            output_name: The name (ID) of this specific output, to be referred to
                with ${{ Validation.<output_name> }}. This does support chaining of the
                namespace to create a more dictionary-like structure
            value: The value that the name will be associated with

        '''
        self.logger.debug(
            'Creating output with id "%s" and value "%s"',
            output_name,
            value,
        )
        if self.driver.name.strip().capitalize() == self._output_browser:
            self._deep_insert('Outputs', output_name, value)
        self._deep_insert('Validation', output_name, value)

    def _deep_insert(self, data_store: str, value_name: str, value: str):
        store = self._context_data[data_store]
        *name_iter, name = value_name.split('.')
        for namespace in name_iter:
            try:
                new_store = store.get(namespace, {})  # type: ignore
                store[namespace] = new_store
                store = new_store
            except Exception:
                raise InvalidOutputName(
                    f'The name \'{value_name}.{name}\' has already been used or the '
                    'namespace is invalid'
                )
        if isinstance(store, str):
            raise InvalidOutputName(
                f'The name \'{value_name}.{name}\' has already been used or the '
                'namespace is invalid'
            )

        store[name] = value  # type: ignore

    def _load_definition_files(self, definition_files: List[str]):
        '''
        Given a list of definition file names, loads all of them into the context data store
        '''
        for definition_file in definition_files:
            self.logger.info(
                'Loading definition file "%s"',
                definition_file
            )
            with open(definition_file) as fp:
                data_dict = json.load(fp)
            self.load_definitions(data_dict)

        self.logger.debug('Final Definition object: "%s"', self._context_data['Definitions'])

    def load_definitions(self, definitions_dict: dict):
        '''
        Loads the given dictionary into the context data, merging the dictionaries and preferring
        the newer configurations wherever there is a conflict

        Args:
            definitions_dict: A dictionary containing all the definitions. Definitions
                are strings saved either in an external file or in the 'definitions'
                object of the validation json that works effectively as a macro, allowing
                test writers to use declarative names for XPaths
        '''
        self._context_data['Definitions'] = pdm.deep_merge(
            self._context_data['Definitions'],
            definitions_dict
        )


def get_default_context(
        plugin_manager: PluginManager,
        debug: bool = False,
        drivers_path: str = '.',
        pretty: bool = False,
        json: str = '',
        is_file: bool = False,
        no_sandbox: bool = False,
        definitions: List[str] = [],
        recreate_context: bool = False,
        logger: Optional[Logger] = None
) -> Context:
    '''
    Gets the default context, creating a new one if necessary.

    If a context object already exists, all of the arguments passed into this function
    are ignored.

    Args:
        plugin_manager: An instance of the plugin manager class to attach to the context
        debug: Whether debug configurations should be enabled
        drivers_path: The directory holding all the required drivers
        pretty: Whether the output json should be pretty-printed
        json: The json data describing the validations
        is_file: Whether the json data was passed in originally raw or
            as a string
        no_sandbox: Specifies that Chrome and Edge should start with the --no-sandbox flag
        definitions: A list file names that contain definitions for Quilla
        recreate_context: Whether a new context object should be created or not
        logger: An optional logger instance. If None, one will be created
            with the NullHandler.

    Returns
        Application context shared for the entire application
    '''
    if Context.default_context is None or recreate_context:
        if logger is not None:
            logger.debug('Creating new default context object')

        Context.default_context = Context(
            plugin_manager,
            debug,
            drivers_path,
            pretty,
            json,
            is_file,
            no_sandbox,
            definitions,
            logger,
        )
    return Context.default_context



'''
A module containing an assortment of utility classes that don't really fit in anywhere else.
These classes define shared behaviour for specific actions that some other classes require, such
as checking for the actual existence of a valid driver when attempting to retrieve the driver,
or alternatively attempting to resolve a valid enum given its type and the value associated with it.
'''


from typing import (
    Type,
    Optional,
    TypeVar
)
from enum import Enum

from selenium.webdriver.remote.webdriver import WebDriver

from quilla.common.exceptions import (
    NoDriverException,
    EnumValueNotFoundException
)


T = TypeVar('T', bound=Enum)


class DriverHolder:
    '''
    Utility class to define shared behaviour for classes that contain
    a driver property
    '''
    def __init__(self, driver: Optional[WebDriver] = None):
        self._driver = driver

    @property
    def driver(self) -> WebDriver:
        '''
        The webdriver attached to this object

        Raises:
            NoDriverException: if the internal driver is currently None
        '''
        if self._driver is None:
            raise NoDriverException
        return self._driver

    @driver.setter
    def driver(self, new_driver: Optional[WebDriver]) -> Optional[WebDriver]:
        self._driver = new_driver

        return new_driver


class EnumResolver:
    '''
    Utility class to define shared behaviour for classes that need to
    resolve string values into appropriate enums
    '''

    # ctx type omitted due to circular import
    @classmethod
    def _name_to_enum(cls, name: str, enum: Type[T], ctx=None) -> T:
        '''
        Converts a string value into the appropriate enum type.
        Useful for inner representations of the data so we're not just working with strings
        everywhere

        Args:
            ctx: the runtime context of the application
            name: the string value to resolve into an enum
            enum: the parent Enum class that contains an entry matching the name argument

        Raises:
            EnumValueNotFoundException: if this resolver fails to resolve an appropriate
                enum value
        '''
        for enum_obj in enum:
            if enum_obj.value == name:
                return enum_obj

        if ctx is not None:
            resolved_plugin_value = ctx.pm.hook.quilla_resolve_enum_from_name(name=name, enum=enum)

            if resolved_plugin_value is not None:
                return resolved_plugin_value

        raise EnumValueNotFoundException(name, enum)



'''
Hooks related to outputs and reports
'''

from quilla.ctx import Context
from quilla.hookspecs import hookspec
from quilla.reports.report_summary import ReportSummary


@hookspec
def quilla_postvalidate(ctx: Context, reports: ReportSummary):
    '''
    A hook called immediately after all validations are executed and the full
    ReportSummary is generated

    Args:
        ctx: The runtime context for the application
        reports: An object capturing all generated reports and giving summary data
    '''



from typing import (
    Dict,
    List,
    Optional
)

# from selenium import webdriver
from selenium.webdriver.remote.webdriver import WebDriver

from quilla.ctx import Context
from quilla.browser import drivers
from quilla.common.enums import (
    BrowserTargets,
)
from quilla.steps.steps_aggregator import StepsAggregator
from quilla.reports.base_report import BaseReport
from quilla.common.exceptions import InvalidBrowserStateException


BrowserSelector = Dict[BrowserTargets, WebDriver]


class BrowserValidations:
    '''
    A class that defines the behaviours for validating a set of steps
    for a specific browser target.

    Args:
        ctx: The runtime context for the application
        target: An enum specifying the target browser that is desired
        url_root: The initial url for the browser to navigate to
        steps: An aggregator class that manages all substeps

    Attributes:
        ctx: The runtime context for the application
    '''
    driver_selector: BrowserSelector = {
        BrowserTargets.FIREFOX: drivers.FirefoxBrowser,
        BrowserTargets.CHROME: drivers.ChromeBrowser,
        BrowserTargets.EDGE: drivers.EdgeBrowser,
    }

    def __init__(
        self,
        ctx: Context,
        target: BrowserTargets,
        url_root: str,
        steps: StepsAggregator,
    ) -> None:
        self._target = target
        self._root = url_root
        self._steps = steps
        self._driver: Optional[WebDriver] = None
        self.ctx = ctx

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, v: BrowserTargets):
        if self._driver is not None:
            raise InvalidBrowserStateException(
                'Cannot change browser target while a driver is set'
            )
        self._target = v

    def init(self):
        '''
        Creates the appropriate driver, sets the start URL to the specified root, and
        sets all steps to have the appropriate driver
        '''
        self.ctx.logger.debug('Initializing browser %s', self._target.value)
        driver: WebDriver = self.driver_selector[self._target](self.ctx)
        self._driver = driver
        driver.get(self._root)
        self._steps.driver = driver  # Set the driver for all the steps
        self.ctx.driver = driver

    def run_steps(self) -> List[BaseReport]:
        '''
        Executes all stored steps. Pass through method for the steps run_steps method

        Returns:
            A list of reports generated by the steps
        '''
        return self._steps.run_steps()

    def clean(self):
        '''
        Closes the browser instance and resets all the step drivers to None state
        '''
        self.ctx.logger.debug('Cleaning up finished browser')
        if self.ctx.close_browser:
            try:
                self._driver.close()
            except Exception as e:
                self.ctx.logger.debug('Could not close browser due to %e', e, exc_info=True)
                pass  # Browser unable to be closed or is already closed
        self._driver = None
        self._steps.driver = None
        self.ctx.driver = None

    def validate(self) -> List[BaseReport]:
        '''
        Initializes the browser, runs the execution steps, and closes up the browser while
        ensuring that any exceptions still allow for cleanup of the browser

        Returns:
            A list of reports produced by the run_steps function

        Raises:
            Exception: Any exception produced by the run_steps function
        '''
        self.init()
        reports = []
        try:
            reports = self.run_steps()
        except Exception as e:
            # Catch exception and crash the program, but clean up
            # the browser first
            self.ctx.logger.error(
                'Exception %s raised for browser %s while running steps',
                e,
                self._target.value,
                exc_info=True,
            )
            raise e
        finally:
            self.clean()

        return reports



'''
Custom exceptions for the UIValidation module
'''

from enum import Enum
from typing import Type


# Exceptions
class UIValidationException(Exception):
    '''
    Base exception for all UIValidation module exceptions
    '''


class NoDriverException(UIValidationException):
    '''
    Exception for when steps are called to action without being bound to a driver
    '''

    def __init__(self):
        super().__init__("No driver currently bound")


class FailedStepException(UIValidationException):
    '''
    Exception for when there is a failed step in the chain
    '''


class EnumValueNotFoundException(UIValidationException):
    '''
    Exception for when an enum value cannot be found by string
    value
    '''

    def __init__(self, str_value: str, enum: Type[Enum]):
        super().__init__(f'Cannot find {enum} with value {str_value}')


class InvalidContextExpressionException(UIValidationException):
    '''
    Exception caused by the context expression syntax being invalid
    '''


class InvalidOutputName(UIValidationException):
    '''
    Exception caused by an invalid output name
    '''


class InvalidBrowserStateException(UIValidationException):
    '''
    Exception caused by attempting to change the browser target
    while the browser is currently open
    '''



'''
This file is used to generate the parser for the documentation. It initializes a plugin manager,
and runs the necessary hooks to set up the parser. This is done so that built-in plugins will also
appear in the usage documentation from ``doc/usage.rst``.
'''

import logging

from quilla import (
    make_parser,
)
from quilla.plugins import get_plugin_manager


def _setup_docs_parser():
    '''
    This function is deliberately not called anywhere, and will not show up
    in the documentation.
    '''
    logger = logging.getLogger(__name__)
    logger.addHandler(logging.NullHandler())

    parser = make_parser()
    pm = get_plugin_manager('.', logger)

    pm.hook.quilla_addopts(parser=parser)

    return parser



from typing import Union

from quilla.common.enums import (
    UITestActions,
    ReportType,
)
from quilla.reports.base_report import BaseReport


class StepFailureReport(BaseReport):
    '''
    Data class for specifying the cause of a step failing in a critical, usually unexpected way.

    Args:
        exception: Either an actual exception class, or a string that describes the failure
        browser: The name of the browser this action failed on
        action: An enum describing the action that was taken
        step_index: The index of the step that failed, for debugging purposes

    Attributes:
        index: The index of the step that failed, for debugging purposes
    '''
    def __init__(
        self,
        exception: Union[Exception, str],
        browser: str,
        action: UITestActions,
        step_index: int
    ):
        super().__init__(ReportType.STEP_FAILURE, browser, action, repr(exception))
        self.index = step_index

    def to_dict(self):
        '''
        Converts this report into a dictionary object

        Returns:
            a dictionary containing the representation of this object
        '''
        return {
            'stepFailureReport': {
                'action': self.action.value,
                'targetBrowser': self.browser,
                'passed': False,
                'stepIndex': self.index,
                'msg': self.msg,
            }
        }

    @classmethod
    def from_dict(cls, report_dict):
        '''
        Converts a dictionary representing this object into a proper StepFailureReport object

        Returns:
            the appropriate StepFailureReport object
        '''
        report = report_dict['stepFailureReport']
        return StepFailureReport(
            report['msg'],
            report['targetBrowser'],
            cls._name_to_enum(report['action'], UITestActions),
            report['stepIndex']
        )



from typing import (
    List,
    Optional,
    Dict,
    Type,
)

from selenium.webdriver.remote.webdriver import WebDriver

from quilla.ctx import Context
from quilla.common.utils import DriverHolder
from quilla.common.enums import (
    UITestActions
)
from quilla.steps.base_steps import (
    BaseStep,
    BaseStepFactory,
)
from quilla.steps.steps import TestStep
from quilla.steps.validations import Validation
from quilla.steps.outputs import OutputValueStep
from quilla.reports import (
    BaseReport,
    StepFailureReport
)


class StepsAggregator(DriverHolder):
    '''
    Test step aggregator interface. Useful for abstracting operations
    done on all the steps.
    '''
    def __init__(
        self,
        ctx: Context,
        steps: List[Dict] = [],
        driver: Optional[WebDriver] = None
    ):
        '''
        Turns an array of dictionaries into appropriate step objects, and saves them in a list
        '''

        self._steps: List[BaseStep] = []
        self._driver = driver
        self.ctx = ctx

        step_factory_selector: Dict[UITestActions, Type[BaseStepFactory]] = {
            UITestActions.VALIDATE: Validation,
            UITestActions.OUTPUT_VALUE: OutputValueStep,
        }

        # Allow plugins to add selectors
        ctx.logger.info('Running "quilla_step_factory_selector" hook')
        ctx.pm.hook.quilla_step_factory_selector(selector=step_factory_selector)

        for step in steps:
            step_factory = step_factory_selector.get(step['action'], TestStep)

            self._steps.append(step_factory.from_dict(ctx, step, driver=driver))  # type: ignore

    @property
    def driver(self) -> WebDriver:
        '''
        The webdriver attached to this object. Setting this property will also set
        the driver of all steps in the aggregator.

        Raises:
            NoDriverException: if the internal driver is currently None
        '''

        return super().driver

    @driver.setter
    def driver(self, new_driver: Optional[WebDriver]):
        for step in self._steps:
            step.driver = new_driver

        self._driver = new_driver

    def run_steps(self) -> List[BaseReport]:
        '''
        Performs all bound steps, collecting generated reports and errors

        Returns:
            A list of reports generated
        '''
        reports: List[BaseReport] = []
        for i, step in enumerate(self._steps):
            try:
                self.ctx.logger.debug('Running step %s', step.action.value)
                report = step.perform()
            except Exception as e:
                if not self.ctx.suppress_exceptions:
                    # If debugging, don't produce reports since a stack trace will be better
                    raise e

                self.ctx.logger.debug(
                    'Performing step %s caused exception %s',
                    step.action.value,
                    e,
                    exc_info=True
                )
                report = StepFailureReport(e, self.driver.name, step.action, i)
                reports.append(report)
                # Exit early, since steps producing exception can prevent future steps from working
                return reports

            if report is not None:
                reports.append(report)

        return reports

    def copy(self) -> "StepsAggregator":
        '''
        Creates a copy of the StepsAggregator object

        This is used so that each browser can have an independent copy of
        the steps, in case any script would want to edit individual browser
        steps
        '''
        steps = []

        for step in self._steps:
            steps.append(step.copy())

        duplicate = StepsAggregator(self.ctx)
        duplicate._steps = steps
        duplicate._driver = self._driver

        return duplicate



from setuptools import setup, find_packages


with open('VERSION') as f:
    version = f.read().strip()

with open('README.md') as f:
    long_description = f.read()

extra_dependencies = {
    'tests': [
        'flake8',
        'mypy',
        'pytest',
        'pytest-cov',
        'pytest-sugar',  # Added for better outputs
        'pytest-xdist',  # Parallelize the tests
    ],
    'docs': [
        'sphinx',
        'sphinx-rtd-theme',
        'sphinx_autodoc_typehints',
        'myst_parser',
        'sphinx_argparse_cli',
    ],
    'pytest': [  # For the plugin
        'pytest'
    ],
    'dev': [
        'pre-commit',
        'types-setuptools',  # Adds typing stubs
    ],
    'release': [
        'wheel',
        'twine',
        'gitchangelog',
        'pystache',
    ]
}

all_dependencies = []

for _, dependencies in extra_dependencies.items():
    all_dependencies.extend(dependencies)

all_dependencies = list(set(all_dependencies))  # Convert to set to remove overlaps
extra_dependencies['all'] = all_dependencies

setup(
    name='quilla',
    version=version,
    description='Declarative UI testing with JSON',
    author='Natalia Maximo',
    author_email='tal.afp.max@gmail.com',
    maintainer='CRE Avengers',
    maintainer_email='cre-avengers@microsoft.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/microsoft/quilla',
    python_requires='>=3.8',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={
        'quilla': ['py.typed']
    },
    include_package_data=True,
    install_requires=[
        'selenium',
        'pluggy',
        'msedge-selenium-tools',
        'pydeepmerge'
    ],
    tests_require=extra_dependencies['tests'],
    extras_require=extra_dependencies,
    entry_points={
        'console_scripts': ['quilla = quilla:run'],
        'pytest11': [
            'quilla = pytest_quilla'
        ]
    },
    project_urls={
        'Issues': 'https://github.com/microsoft/quilla/issues',
        'Discussions': 'https://github.com/microsoft/quilla/discussions'
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Typing :: Typed',
    ]
)



'''
Module with all requried enums for the UIValidation
'''

from enum import Enum


# Enums
class ValidationTypes(Enum):
    '''
    The currently-supported types of validation allowed
    '''
    XPATH = 'XPath'
    URL = 'URL'


class ValidationStates(Enum):
    '''
    Base class for validation state enums
    '''


class XPathValidationStates(ValidationStates):
    '''
    States that the XPath validation class recognizes
    '''
    EXISTS = 'Exists'
    NOT_EXISTS = 'NotExists'
    VISIBLE = 'Visible'
    NOT_VISIBLE = 'NotVisible'
    TEXT_MATCHES = 'TextMatches'
    NOT_TEXT_MATCHES = 'NotTextMatches'
    HAS_PROPERTY = 'HasProperty'
    NOT_HAS_PROPERTY = 'NotHasProperty'
    PROPERTY_HAS_VALUE = 'PropertyHasValue'
    NOT_PROPERTY_HAS_VALUE = 'NotPropertyHasValue'
    HAS_ATTRIBUTE = 'HasAttribute'
    NOT_HAS_ATTRIBUTE = 'NotHasAttribute'
    ATTRIBUTE_HAS_VALUE = 'AttributeHasValue'
    NOT_ATTRIBUTE_HAS_VALUE = 'NotAttributeHasValue'


class URLValidationStates(ValidationStates):
    '''
    States that the URL validation class recognizes
    '''
    CONTAINS = 'Contains'
    NOT_CONTAINS = 'NotContains'
    EQUALS = 'Equals'
    NOT_EQUALS = 'NotEquals'
    MATCHES = 'Matches'
    NOT_MATCHES = 'NotMatches'


class UITestActions(Enum):
    '''
    All supported UI test actions
    '''
    CLICK = 'Click'
    CLEAR = 'Clear'
    SEND_KEYS = 'SendKeys'
    WAIT_FOR_EXISTENCE = 'WaitForExistence'
    WAIT_FOR_VISIBILITY = 'WaitForVisibility'
    NAVIGATE_TO = 'NavigateTo'
    VALIDATE = 'Validate'
    REFRESH = "Refresh"
    ADD_COOKIES = "AddCookies"
    SET_COOKIES = "SetCookies"
    REMOVE_COOKIE = "RemoveCookie"
    CLEAR_COOKIES = "ClearCookies"
    NAVIGATE_FORWARD = "NavigateForward"
    NAVIGATE_BACK = "NavigateBack"
    SET_BROWSER_SIZE = "SetBrowserSize"
    HOVER = "Hover"
    OUTPUT_VALUE = "OutputValue"


class ReportType(Enum):
    '''
    All the currently supported report types
    '''
    VALIDATION = "Validation"
    STEP_FAILURE = "StepFailure"


class BrowserTargets(Enum):
    '''
    All the currently supported browser targets
    '''
    FIREFOX = "Firefox"
    CHROME = "Chrome"
    EDGE = "Edge"


class OutputSources(Enum):
    '''
    Supported sources for the OutputValue action
    '''
    LITERAL = 'Literal'
    XPATH_TEXT = 'XPathText'
    XPATH_PROPERTY = 'XPathProperty'



'''
Hooks that are related to configuration, such as logger configs, parser additions,
etc.
'''
from enum import Enum
from logging import Logger
from argparse import (
    ArgumentParser,
    Namespace
)
from typing import (
    Type,
    TypeVar,
    Optional,
)

from quilla.hookspecs import hookspec
from quilla.ctx import Context
from quilla.ui_validation import QuillaTest


T = TypeVar('T', bound=Enum)


@hookspec
def quilla_configure_logger(logger: Logger):
    '''
    A hook called immediately after the plugin manager is created. This is the very
    first hook called, and allows plugins to register additional handlers, formatters,
    or otherwise modify the logger used throughout Quilla. Note, the Context is not
    yet created

    To help in debugging, it is recommended that plugins register their own StreamHandler
    to the logger with a filter that shows only the messages relevant to the plugin.

    Args:
        logger: The configured logger instance for Quilla
    '''


@hookspec
def quilla_addopts(parser: ArgumentParser):
    '''
    A hook to allow plugins to add additional arguments to the argument parser.
    This can be used if a plugin requires additional parameters or data in some way.

    This is called after the initial argument parser setup

    Args:
        parser: The argparse Argument Parser instance used by the application
    '''


@hookspec
def quilla_configure(ctx: Context, args: Namespace):
    '''
    A hook to allow plugins to modify the context object, either changing its data
    or adding data to it.

    This is called after the initial setup of the context object

    Args:
        ctx: The runtime context for the application
        args: Parsed CLI args, in case they are needed
    '''


@hookspec
def quilla_prevalidate(validation: QuillaTest):
    '''
    A hook called immediately before the validations attempt to be resolved
    (i.e. before `validations.validate_all()` is called)

    Args:
        validation: The collected validations from the json passed to
            the application
    '''


@hookspec(firstresult=True)
def quilla_resolve_enum_from_name(name: str, enum: Type[T]) -> Optional[T]:
    '''
    A hook called when a value specified by the quilla test should be resolved to an
    enum, but no enum has been found. This is to allow plugins to register custom
    enum values for quilla, such as new step actions, validation types, validation states,
    output sources, etc.

    Args:
        name: the string value specified in the quilla test
        enum: The enum subclass type that is being attempted to be resolved. This
            should give an indication as to what is being resolved. For example,
            UITestActions is the enum type being resolved for the 'actions' field.

    Returns:
        The resolved enum, if it can be resolved. None if the plugin can't resolve
        the value.
    '''



from typing import (
    Optional,
    Callable,
    Dict,
    Any,
)
from abc import abstractclassmethod, abstractmethod

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.by import By

from quilla.ctx import Context
from quilla.reports import (
    BaseReport,
    ValidationReport
)
from quilla.common.utils import (
    DriverHolder,
    EnumResolver
)
from quilla.common.enums import (
    UITestActions,
    ValidationTypes,
    ValidationStates,
)
from quilla.common.exceptions import FailedStepException


class BaseStep(DriverHolder, EnumResolver):
    '''
    Base class for all step objects

    Args:
        ctx: The runtime context for the application
        action_type: Enum defining which of the supported actions this class represents
        driver: An optional argument to allow the driver to be bound at object creation.

    Attributes:
        ctx: The runtime context for the application
        action: Enum defining which of the supported actions this class represents
    '''
    def __init__(
        self,
        ctx: Context,
        action_type: UITestActions,
        target: Optional[str] = None,
        parameters: Optional[dict] = None,
        driver: Optional[WebDriver] = None
    ):
        self.action = action_type
        self.ctx = ctx
        self.target = target
        self.parameters = parameters
        super().__init__(driver)

    @abstractmethod
    def perform(self) -> Optional[BaseReport]:  # pragma: no cover
        '''
        Runs the necessary action. If the action is a Validate action, will return a
        ValidationReport

        Returns:
            A report produced by the step, or None if no report is required
        '''
        pass

    @abstractmethod
    def copy(self) -> "BaseStep":
        '''
        Returns a copy of the current Step object
        '''

    @property
    def target(self):
        '''
        The target for an action, if applicable. Will resolve all context
        expressions before being returned
        '''
        if self._target is not None:
            return self.ctx.perform_replacements(self._target)

    @target.setter
    def target(self, val: str) -> str:
        self._target = val
        return val

    @property
    def parameters(self):
        '''
        Parameters for an action, if applicable. Will resolve all context
        expressions before being returned
        '''
        if self._parameters is not None:
            return self._deep_replace(self.ctx, self._parameters.copy())

    @parameters.setter
    def parameters(self, val: dict) -> dict:
        self._parameters = val
        return val

    def _deep_replace(self, ctx: Context, params: Dict[str, Any]):
        for key, value in params.items():
            if isinstance(value, str):
                params[key] = ctx.perform_replacements(value)
            elif isinstance(value, Dict):
                params[key] = self._deep_replace(ctx, value)
            # (TODO): Add list support to deep replacement
        return params

    def _verify_parameters(self, *parameters: str):
        for parameter in parameters:
            if parameter not in self.parameters:  # type: ignore
                raise FailedStepException(
                    f'"{parameter}" parameter not specified for "{self.action.value}" action'
                )

    def _verify_target(self):
        if self.target is None:
            raise FailedStepException(f'No specified target for "{self.action.value}" action')

    @property
    def element(self) -> WebElement:
        '''
        Located WebElement instance
        '''
        return self.driver.find_element(*self.locator)

    @property
    def locator(self):
        '''
        Locator for selenium to find web elements
        '''
        return (By.XPATH, self.target)


class BaseStepFactory:
    @abstractclassmethod
    def from_dict(cls, ctx: Context, step: Dict, driver: Optional[WebDriver] = None) -> BaseStep:
        '''
        Given a context, step dictionary, and optionally a driver, return an appropriate subclass
        of BaseStep

        Args:
            ctx: The runtime context of the application
            step: A dictionary defining the step
            driver: Optionally, a driver to bind to the resulting step

        Returns:
            The resulting step object
        '''


class BaseValidation(BaseStep):
    '''
    Base validation class with shared functionality for all validations

    Args:
        ctx: The runtime context for the application
        type_: An enum describing the type of supported validation
        target: The general target that this validation will seek. What it means is
            specific to each subclass of this class
        state: An enum describing the desired state of this validation. The specific
            enum is a subclass of the ValidationStates class specific to the subclass
            of BaseValidation being used
        selector: A dictionary that maps the state enum to the appropriate function
    '''
    def __init__(
        self,
        ctx: Context,
        type_: ValidationTypes,
        target: str,
        state: ValidationStates,
        selector: Dict[ValidationStates, Callable[[], ValidationReport]],
        parameters: Optional[Dict],
        driver: Optional[WebDriver] = None,
    ) -> None:
        super().__init__(
            ctx,
            UITestActions.VALIDATE,
            target=target,
            parameters=parameters,
            driver=driver
        )
        self._type = type_
        self._state = state
        self._driver = driver
        self._selector = selector
        self._report: Optional[ValidationReport] = None

    def copy(self) -> "BaseValidation":
        # All classes derived from BaseValidation only need these parameters
        return self.__class__(  # type: ignore
            self.ctx,  # type: ignore
            self._target,  # type: ignore
            self._state,  # type: ignore
            self._parameters,  # type: ignore
            self._driver  # type: ignore
        )

    def perform(self) -> ValidationReport:
        '''
        Performs the correct action based on what is defined within the selector,
        and returns the resulting report produced.

        Returns:
            A report summarizing the results of the executed validation
        '''
        self.ctx.logger.debug(
            'Performing %s with desired state "%s"',
            self._type.value,
            self._state.value
        )
        action_function = self._selector[self._state]
        self._report = action_function()

        return self._report

    def _create_report(self, success: bool, msg: str = "") -> ValidationReport:
        '''
        Creates a new validation report. Used to simplify the common shared
        behaviour that all validation reports require

        Args:
            success: Value representing the successfulness of the validation. True if
                the validation passed, False otherwise
            msg: An optional string message to be included in the report

        Returns:
            A report summarizing the results of the executed validation
        '''
        return ValidationReport(
            self._type.value,
            self._target,
            self._state.value,
            self.driver.name.capitalize(),
            success=success,
            msg=msg
        )
